/*
 * Copyright (C) 2020 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.server.integrity.serializer;

import java.io.IOException;
import java.io.OutputStream;

/**
 * An output stream that tracks the total number written bytes since construction and allows
 * querying this value any time during the execution.
 *
 * This class is used for constructing the rule indexing.
 */
public class ByteTrackedOutputStream {

    private static long sWrittenBytesCount;
    private static OutputStream sOutputStream;

    public ByteTrackedOutputStream(OutputStream outputStream) {
        sWrittenBytesCount = 0;
        sOutputStream = outputStream;
    }

    /**
     * Writes the given bytes into the output stream provided in constructor and updates the
     * total number of written bytes.
     */
    public void write(byte[] bytes) throws IOException {
        sWrittenBytesCount += bytes.length;
        sOutputStream.write(bytes);
    }

    /**
     * Returns the total number of bytes written into the output stream at the requested time.
     */
    public long getWrittenBytesCount() {
        return sWrittenBytesCount;
    }
}
